@@ -17,6 +17,7 @@
 OpenStack Swift client library used internally
 """
 import socket
+import re
 import requests
 import logging
 import warnings
@@ -38,6 +39,7 @@ from swiftclient.utils import (
 # Default is 100, increase to 256
 http_client._MAXHEADERS = 256
 
+VERSIONFUL_AUTH_PATH = re.compile('v[2-3](?:\.0)?$')
 AUTH_VERSIONS_V1 = ('1.0', '1', 1)
 AUTH_VERSIONS_V2 = ('2.0', '2', 2)
 AUTH_VERSIONS_V3 = ('3.0', '3', 3)
@@ -149,7 +151,7 @@ def http_log(args, kwargs, resp, body):
         elif element in ('GET', 'POST', 'PUT'):
             string_parts.append(' -X %s' % element)
         else:
-            string_parts.append(' %s' % element)
+            string_parts.append(' %s' % parse_header_string(element))
     if 'headers' in kwargs:
         headers = scrub_headers(kwargs['headers'])
         for element in headers:
@@ -271,6 +273,9 @@ class _ObjectBody(object):
     def __next__(self):
         return self.next()
 
+    def close(self):
+        self.resp.close()
+
 
 class _RetryBody(_ObjectBody):
     """
@@ -302,7 +307,7 @@ class _RetryBody(_ObjectBody):
         self.obj = obj
         self.query_string = query_string
         self.response_dict = response_dict
-        self.headers = headers if headers is not None else {}
+        self.headers = dict(headers) if headers is not None else {}
         self.bytes_read = 0
 
     def read(self, length=None):
@@ -453,11 +458,23 @@ class HTTPConnection(object):
         self.resp.status = self.resp.status_code
         old_getheader = self.resp.raw.getheader
 
+        def _decode_header(string):
+            if string is None or six.PY2:
+                return string
+            return string.encode('iso-8859-1').decode('utf-8')
+
+        def _encode_header(string):
+            if string is None or six.PY2:
+                return string
+            return string.encode('utf-8').decode('iso-8859-1')
+
         def getheaders():
-            return self.resp.headers.items()
+            return [(_decode_header(k), _decode_header(v))
+                    for k, v in self.resp.headers.items()]
 
         def getheader(k, v=None):
-            return old_getheader(k.lower(), v)
+            return _decode_header(old_getheader(
+                _encode_header(k.lower()), _encode_header(v)))
 
         def releasing_read(*args, **kwargs):
             chunk = self.resp.raw.read(*args, **kwargs)
@@ -511,8 +528,9 @@ def get_auth_1_0(url, user, key, snet, **kwargs):
         netloc = parsed[1]
         parsed[1] = 'snet-' + netloc
         url = urlunparse(parsed)
-    return url, resp.getheader('x-storage-token',
-                               resp.getheader('x-auth-token'))
+
+    token = resp.getheader('x-storage-token', resp.getheader('x-auth-token'))
+    return url, token
 
 
 def get_keystoneclient_2_0(auth_url, user, key, os_options, **kwargs):
@@ -555,7 +573,10 @@ def get_auth_keystone(auth_url, user, key, os_options, **kwargs):
 
     # Add the version suffix in case of versionless Keystone endpoints. If
     # auth_version is also unset it is likely that it is v3
-    if len(urlparse(auth_url).path) <= 1:
+    if not VERSIONFUL_AUTH_PATH.match(
+            urlparse(auth_url).path.rstrip('/').rsplit('/', 1)[-1]):
+        # Normalize auth_url to end in a slash because urljoin
+        auth_url = auth_url.rstrip('/') + '/'
         if auth_version and auth_version in AUTH_VERSIONS_V2:
             auth_url = urljoin(auth_url, "v2.0")
         else:
@@ -565,7 +586,7 @@ def get_auth_keystone(auth_url, user, key, os_options, **kwargs):
 
     # Legacy default if not set
     if auth_version is None:
-        auth_version = 'v2.0'
+        auth_version = '2'
     ksclient, exceptions = _import_keystone_client(auth_version)
 
     try:
@@ -644,8 +665,10 @@ def get_auth(auth_url, user, key, **kwargs):
     if session:
         service_type = os_options.get('service_type', 'object-store')
         interface = os_options.get('endpoint_type', 'public')
+        region_name = os_options.get('region_name')
         storage_url = session.get_endpoint(service_type=service_type,
-                                           interface=interface)
+                                           interface=interface,
+                                           region_name=region_name)
         token = session.get_token()
     elif auth_version in AUTH_VERSIONS_V1:
         storage_url, token = get_auth_1_0(auth_url,
@@ -847,13 +870,15 @@ def post_account(url, token, headers, http_conn=None, response_dict=None,
     path = parsed.path
     if query_string:
         path += '?' + query_string
-    headers['X-Auth-Token'] = token
+    req_headers = {'X-Auth-Token': token}
     if service_token:
-        headers['X-Service-Token'] = service_token
-    conn.request(method, path, data, headers)
+        req_headers['X-Service-Token'] = service_token
+    if headers:
+        req_headers.update(headers)
+    conn.request(method, path, data, req_headers)
     resp = conn.getresponse()
     body = resp.read()
-    http_log((url, method,), {'headers': headers}, resp, body)
+    http_log((url, method,), {'headers': req_headers}, resp, body)
 
     store_response(resp, response_dict)
 
@@ -895,12 +920,6 @@ def get_container(url, token, container, marker=None, limit=None,
     """
     if not http_conn:
         http_conn = http_connection(url)
-    if headers:
-        headers = dict(headers)
-    else:
-        headers = {}
-    headers['X-Auth-Token'] = token
-    headers['Accept-Encoding'] = 'gzip'
     if full_listing:
         rv = get_container(url, token, container, marker, limit, prefix,
                            delimiter, end_marker, path, http_conn,
@@ -935,17 +954,20 @@ def get_container(url, token, container, marker=None, limit=None,
         qs += '&path=%s' % quote(path)
     if query_string:
         qs += '&%s' % query_string.lstrip('?')
+    req_headers = {'X-Auth-Token': token, 'Accept-Encoding': 'gzip'}
     if service_token:
-        headers['X-Service-Token'] = service_token
+        req_headers['X-Service-Token'] = service_token
+    if headers:
+        req_headers.update(headers)
     method = 'GET'
-    conn.request(method, '%s?%s' % (cont_path, qs), '', headers)
+    conn.request(method, '%s?%s' % (cont_path, qs), '', req_headers)
     resp = conn.getresponse()
     body = resp.read()
     http_log(('%(url)s%(cont_path)s?%(qs)s' %
               {'url': url.replace(parsed.path, ''),
                'cont_path': cont_path,
                'qs': qs}, method,),
-             {'headers': headers}, resp, body)
+             {'headers': req_headers}, resp, body)
 
     if resp.status < 200 or resp.status >= 300:
         raise ClientException.from_response(resp, 'Container GET failed', body)
@@ -1018,23 +1040,23 @@ def put_container(url, token, container, headers=None, http_conn=None,
         parsed, conn = http_connection(url)
     path = '%s/%s' % (parsed.path, quote(container))
     method = 'PUT'
-    if not headers:
-        headers = {}
-    headers['X-Auth-Token'] = token
+    req_headers = {'X-Auth-Token': token}
     if service_token:
-        headers['X-Service-Token'] = service_token
-    if 'content-length' not in (k.lower() for k in headers):
-        headers['Content-Length'] = '0'
+        req_headers['X-Service-Token'] = service_token
+    if headers:
+        req_headers.update(headers)
+    if 'content-length' not in (k.lower() for k in req_headers):
+        req_headers['Content-Length'] = '0'
     if query_string:
         path += '?' + query_string.lstrip('?')
-    conn.request(method, path, '', headers)
+    conn.request(method, path, '', req_headers)
     resp = conn.getresponse()
     body = resp.read()
 
     store_response(resp, response_dict)
 
     http_log(('%s%s' % (url.replace(parsed.path, ''), path), method,),
-             {'headers': headers}, resp, body)
+             {'headers': req_headers}, resp, body)
     if resp.status < 200 or resp.status >= 300:
         raise ClientException.from_response(resp, 'Container PUT failed', body)
 
@@ -1061,16 +1083,18 @@ def post_container(url, token, container, headers, http_conn=None,
         parsed, conn = http_connection(url)
     path = '%s/%s' % (parsed.path, quote(container))
     method = 'POST'
-    headers['X-Auth-Token'] = token
+    req_headers = {'X-Auth-Token': token}
     if service_token:
-        headers['X-Service-Token'] = service_token
+        req_headers['X-Service-Token'] = service_token
+    if headers:
+        req_headers.update(headers)
     if 'content-length' not in (k.lower() for k in headers):
-        headers['Content-Length'] = '0'
-    conn.request(method, path, '', headers)
+        req_headers['Content-Length'] = '0'
+    conn.request(method, path, '', req_headers)
     resp = conn.getresponse()
     body = resp.read()
     http_log(('%s%s' % (url.replace(parsed.path, ''), path), method,),
-             {'headers': headers}, resp, body)
+             {'headers': req_headers}, resp, body)
 
     store_response(resp, response_dict)
 
@@ -1188,7 +1212,7 @@ def get_object(url, token, container, name, http_conn=None,
 
 
 def head_object(url, token, container, name, http_conn=None,
-                service_token=None, headers=None):
+                service_token=None, headers=None, query_string=None):
     """
     Get object info
 
@@ -1209,6 +1233,8 @@ def head_object(url, token, container, name, http_conn=None,
     else:
         parsed, conn = http_connection(url)
     path = '%s/%s/%s' % (parsed.path, quote(container), quote(name))
+    if query_string:
+        path += '?' + query_string
     if headers:
         headers = dict(headers)
     else:
@@ -1365,14 +1391,16 @@ def post_object(url, token, container, name, headers, http_conn=None,
     else:
         parsed, conn = http_connection(url)
     path = '%s/%s/%s' % (parsed.path, quote(container), quote(name))
-    headers['X-Auth-Token'] = token
+    req_headers = {'X-Auth-Token': token}
     if service_token:
-        headers['X-Service-Token'] = service_token
-    conn.request('POST', path, '', headers)
+        req_headers['X-Service-Token'] = service_token
+    if headers:
+        req_headers.update(headers)
+    conn.request('POST', path, '', req_headers)
     resp = conn.getresponse()
     body = resp.read()
     http_log(('%s%s' % (url.replace(parsed.path, ''), path), 'POST',),
-             {'headers': headers}, resp, body)
+             {'headers': req_headers}, resp, body)
 
     store_response(resp, response_dict)
 
@@ -1540,7 +1568,7 @@ class Connection(object):
                  os_options=None, auth_version="1", cacert=None,
                  insecure=False, cert=None, cert_key=None,
                  ssl_compression=True, retry_on_ratelimit=False,
-                 timeout=None, session=None):
+                 timeout=None, session=None, force_auth_retry=False):
         """
         :param authurl: authentication URL
         :param user: user name to authenticate as
@@ -1576,6 +1604,8 @@ class Connection(object):
                                    after a backoff.
         :param timeout: The connect timeout for the HTTP connection.
         :param session: A keystoneauth session object.
+        :param force_auth_retry: reset auth info even if client got unexpected
+                                 error except 401 Unauthorized.
         """
         self.session = session
         self.authurl = authurl
@@ -1608,6 +1638,7 @@ class Connection(object):
         self.auth_end_time = 0
         self.retry_on_ratelimit = retry_on_ratelimit
         self.timeout = timeout
+        self.force_auth_retry = force_auth_retry
 
     def close(self):
         if (self.http_conn and isinstance(self.http_conn, tuple)
@@ -1722,6 +1753,10 @@ class Connection(object):
                     pass
                 else:
                     raise
+
+            if self.force_auth_retry:
+                self.url = self.token = self.service_token = None
+
             sleep(backoff)
             backoff = min(backoff * 2, self.max_backoff)
             if reset_func:
@@ -1785,9 +1820,10 @@ class Connection(object):
                            query_string=query_string,
                            headers=headers)
 
-    def head_object(self, container, obj, headers=None):
+    def head_object(self, container, obj, headers=None, query_string=None):
         """Wrapper for :func:`head_object`"""
-        return self._retry(None, head_object, container, obj, headers=headers)
+        return self._retry(None, head_object, container, obj, headers=headers,
+                           query_string=query_string)
 
     def get_object(self, container, obj, resp_chunk_size=None,
                    query_string=None, response_dict=None, headers=None):
